;; Generated package description from geiser-guile.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "geiser-guile" "0.28.5" "Guile and Geiser talk to each other" '((emacs "27.1") (transient "0.3") (geiser "0.33")) :commit "cbab81bd2dcb4c787bcda4ae18062db3087e6887" :authors '(("Jose Antonio Ortega Ruiz" . "(jao@gnu.org)")) :maintainer '("Jose Antonio Ortega Ruiz" . "(jao@gnu.org)") :keywords '("languages" "guile" "scheme" "geiser") :url "https://gitlab.com/emacs-geiser/guile")
